@@ -1,13 +1,13 @@
 /* dd_rescue.c */
 /* 
- * dd_rescue copies your data from one file to another
- * files might as well be block devices, such as hd partitions
- * unlike dd, it does not necessarily abort on errors, but
+ * dd_rescue copies your data from one file to another.
+ * Files might as well be block devices, such as hd partitions.
+ * Unlike dd, it does not necessarily abort on errors but
  * continues to copy the disk, possibly leaving holes behind.
  * Also, it does NOT truncate the output file, so you can copy 
  * more and more pieces of your data, as time goes by.
- * So, this tool is suitable for rescueing data of crashed disk,
- * and that's the reason, it has been written by me.
+ * This tool is thus suitable for rescueing data of crashed disk,
+ * and that's the reason it has been written by me.
  *
  * (c) Kurt Garloff <garloff@suse.de>, 11/97, 10/99
  * Copyright: GNU GPL
@@ -20,13 +20,18 @@
  * TODO:
  * - Use termcap to fetch cursor up/down codes
  * - Better handling of write errors: also try sub blocks
+ * - Optional colors
+ * - Use dlopen to open libfallocate rather than linking to it ...
  */
 
 #ifndef VERSION
 # define VERSION "(unknown)"
 #endif
+#ifndef __COMPILER__
+# define "(unknown compiler)"
+#endif
 
-#define ID "$Id: dd_rescue.c,v 1.52 2006/07/23 11:47:50 garloff Exp $"
+#define ID "$Id: dd_rescue.c,v 1.112 2010/10/11 09:50:32 garloff Exp $"
 
 #ifndef SOFTBLOCKSIZE
 # define SOFTBLOCKSIZE 65536
@@ -46,24 +51,52 @@
 #include <unistd.h>
 #include <string.h>
 #include <ctype.h>
-#include <getopt.h>
-#include <fcntl.h>
 #include <errno.h>
 #include <signal.h>
 #include <time.h>
 #include <utime.h>
 #include <sys/time.h>
 #include <sys/stat.h>
+#ifdef HAVE_GETOPT_H
+#include <getopt.h>
+#endif
+// hack around buggy splice definition(!)
+#define splice oldsplice
+#include <fcntl.h>
+#undef splice
+
+#ifdef HAVE_LIBFALLOCATE
+#include <fallocate.h>
+#endif
+
+/* splice */
+#ifdef __linux__
+# define __KERNEL__
+# include <asm/unistd.h>
+# ifdef __NR_splice
+#  define HAVE_SPLICE 1
+#  if 1
+static inline long splice(int fdin, loff_t *off_in, int fdout, 
+			      loff_t *off_out, size_t len, unsigned int flags)
+{
+	return syscall(__NR_splice, fdin, off_in, fdout, off_out, len, flags);
+}
+#  else
+_syscall6(long, splice, int, fdin, loff_t*, off_in, int, fdout, loff_t*, off_out, size_t, len, unsigned int, flags);
+#  endif
+# endif
+#endif
+
 
 int softbs, hardbs, syncfreq;
 int maxerr, nrerr, reverse, dotrunc, abwrerr, sparse, nosparse;
 int verbose, quiet, interact, force;
-void* buf;
+unsigned char* buf;
 char *lname, *iname, *oname, *bbname = NULL;
-off_t ipos, opos, xfer, lxfer, sxfer, fxfer, maxxfer;
+off_t ipos, opos, xfer, lxfer, sxfer, fxfer, maxxfer, init_opos, ilen, estxfer;
 
-int ides, odes, identical, pres;
-int o_dir_in, o_dir_out;
+int ides, odes, identical, pres, falloc;
+int o_dir_in, o_dir_out, dosplice;
 char i_chr, o_chr;
 
 FILE *logfd;
@@ -71,9 +104,40 @@ struct timeval starttime, lasttime, currenttime;
 struct timezone tz;
 clock_t startclock;
 
-const char* up = "\x1b[A"; //] 
-const char* down = "\n";
-const char* right = "\x1b[C"; //]
+#ifndef UP
+# define UP "\x1b[A"
+# define DOWN "\n"
+# define RIGHT "\x1b[C"
+#endif
+
+const char* up = UP;
+const char* down = DOWN;
+const char* right = RIGHT;
+char *graph;
+
+#ifdef MISS_STRSIGNAL
+static char sbuf[16];
+static char* strsignal(int sig)
+{
+	sprintf(sbuf, "sig %i", sig);
+	return sbuf;
+}
+#endif
+#ifdef MISS_PREAD
+static ssize_t pread(int fd, void *buf, size_t sz, off_t off)
+{
+	if (lseek(fd, off, SEEK_SET))
+		return -1;
+	return read(fd, buf, sz);
+}
+
+static ssize_t pwrite(int fd, void *buf, size_t sz, off_t off)
+{
+	if (lseek(fd, off, SEEK_SET))
+		return -1;
+	return write(fd, buf, sz);
+}
+#endif
 
 inline float difftimetv(const struct timeval* const t2, 
 			const struct timeval* const t1)
@@ -99,7 +163,7 @@ int fplog(FILE* const file, const char * const fmt, ...)
 	return ret;
 }
 
-int check_identical(const char* const in, const char* const on)
+static int check_identical(const char* const in, const char* const on)
 {
 	int err = 0;
 	struct stat istat, ostat;
@@ -119,7 +183,7 @@ int check_identical(const char* const in, const char* const on)
 }
 
 
-int openfile(const char* const fname, const int flags)
+static int openfile(const char* const fname, const int flags)
 {
 	int fdes;
 	if (!strcmp(fname, "-")) {
@@ -138,7 +202,7 @@ int openfile(const char* const fname, const int flags)
 }
 
 /* Checks whether files are seekable */
-void check_seekable(const int id, const int od)
+static void check_seekable(const int id, const int od)
 {
 	errno = 0;
 	if (lseek(id, (off_t)0, SEEK_SET) != 0) {
@@ -157,10 +221,153 @@ void check_seekable(const int id, const int od)
 	errno = 0;
 }
 
+/* Calc position in graph */
+inline int gpos(off_t off)
+{
+	static const int glen = 40; //strlen(graph) - 2;
+	return 1+(glen*off/ilen);
+}
+
+/* Prepare graph */
+static void preparegraph()
+{
+	if (!ilen)
+		return;
+	graph = strdup(":.........................................:");
+	if (reverse) {
+		graph[gpos(ipos)+1] = '<';
+		graph[gpos(ipos-estxfer)-1] = '>';
+
+	} else {
+		graph[gpos(ipos)-1] = '>';
+		graph[gpos(ipos+estxfer)+1] = '<';
+	}
+}
+
+void updgraph(int err)
+{
+	int off;
+	if (!ilen)
+		return;
+	off = gpos(ipos);
+	if (graph[off] == 'x')
+		return;
+	if (err)
+		graph[off] = 'x';
+	else
+		graph[off] = '-';
+}
+
+/* Tries to determine size of input file */
+void input_length()
+{
+	struct stat stbuf;
+	estxfer = maxxfer;
+	if (reverse) {
+		if (ipos)
+			ilen = ipos;
+		else
+			ilen = maxxfer;
+	} else
+		ilen = ipos + maxxfer;
+	if (estxfer)
+		preparegraph();
+	if (i_chr)
+		return;
+	if (fstat(ides, &stbuf))
+		return;
+	if (S_ISLNK(stbuf.st_mode))
+		return;
+	if (S_ISCHR(stbuf.st_mode)) {
+		i_chr = 1;
+		return;
+	}
+	if (S_ISBLK(stbuf.st_mode)) {
+		/* Do magic to figure size of block dev */
+		off_t p = lseek(ides, 0, SEEK_CUR);
+		if (p == -1)
+			return;
+		ilen = lseek(ides, 0, SEEK_END) + 1;
+		lseek(ides, p, SEEK_SET);
+	} else {
+		off_t diff;
+		ilen = stbuf.st_size;
+		if (!ilen)
+			return;
+		diff = ilen - stbuf.st_blocks*512;
+		if (diff >= 4096 && (float)diff/ilen > 0.05 && !quiet)
+		       fplog(stderr, "dd_rescue: (info) %s is sparse (%i%%) %s\n", iname, (int)(100.0*diff/ilen), (sparse? "": ", consider -a"));
+	}
+	if (!ilen)
+		return;
+	if (!reverse)
+		estxfer = ilen - ipos;
+	else
+		estxfer = ipos;
+	if (maxxfer && estxfer > maxxfer)
+		estxfer = maxxfer;
+	if (!quiet)
+		fplog(stderr, "dd_rescue: (info) expect to copy %LikB from %s\n",
+			estxfer/1024, iname);
+	preparegraph();
+}
+
+static void sparse_output_warn()
+{
+	struct stat stbuf;
+	off_t eff_opos;
+	if (o_chr)
+		return;
+	if (fstat(odes, &stbuf))
+		return;
+	if (S_ISCHR(stbuf.st_mode)) {
+		o_chr = 1;
+		return;
+	}
+	if (S_ISBLK(stbuf.st_mode)) {
+		if (sparse || !nosparse)
+			fplog(stderr, "dd_rescue: (warning): %s is a block device; -a not recommended; -A recommended\n", oname);
+		return;
+	}
+	eff_opos = opos == -1? ipos: opos;
+	if (sparse && (eff_opos < stbuf.st_size))
+		fplog(stderr, "dd_rescue: (warning): write into %s (@%li/%li): sparse not recommended\n", 
+				oname, eff_opos, stbuf.st_size);
+}
+
+#ifdef HAVE_FALLOCATE
+static void do_fallocate()
+{
+	struct stat stbuf;
+	off_t to_falloc, alloced;
+	if (o_chr)
+		return;
+	if (!estxfer)
+		return;
+	if (fstat(odes, &stbuf))
+		return;
+	if (!S_ISREG(stbuf.st_mode))
+		return;
+	alloced = stbuf.st_blocks*512 - opos;
+	to_falloc = estxfer - (alloced < 0 ? 0 : alloced);
+	if (to_falloc <= 0)
+		return;
+#ifdef HAVE_LIBFALLOCATE
+	if (linux_fallocate64(odes, FALLOC_FL_KEEP_SIZE, 
+			      opos, to_falloc))
+#else
+	if (fallocate64(odes, 1, opos, to_falloc))
+#endif
+	       fplog(stderr, "dd_rescue: (warning): fallocate %s (%Li, %Li) failed: %s\n",
+			       oname, opos, to_falloc, strerror(errno));
+}
+#endif
+
 
 void doprint(FILE* const file, const int bs, const clock_t cl, 
 	     const float t1, const float t2, const int sync)
 {
+	float avgrate = (float)xfer/(t1*1024);
 	fprintf(file, "dd_rescue: (info): ipos:%12.1fk, opos:%12.1fk, xferd:%12.1fk\n",
 		(float)ipos/1024, (float)opos/1024, (float)xfer/1024);
 	fprintf(file, "             %s  %s  errs:%7i, errxfer:%12.1fk, succxfer:%12.1fk\n",
@@ -169,13 +376,24 @@ void doprint(FILE* const file, const int bs, const clock_t cl,
 	if (sync || (file != stdin && file != stdout) )
 		fprintf(file, "             +curr.rate:%9.0fkB/s, avg.rate:%9.0fkB/s, avg.load:%5.1f%%\n",
 			(float)(xfer-lxfer)/(t2*1024),
-			(float)xfer/(t1*1024),
+			avgrate,
 			100.0*(cl-startclock)/(CLOCKS_PER_SEC*t1));
 	else
 		fprintf(file, "             -curr.rate:%s%s%s%s%s%s%s%s%skB/s, avg.rate:%9.0fkB/s, avg.load:%5.1f%%\n",
 			right, right, right, right, right, right, right, right, right,
-			(float)xfer/(t1*1024),
+			avgrate,
 			100.0*(cl-startclock)/(CLOCKS_PER_SEC*t1));
+	if (estxfer && avgrate > 0) {
+		int sec = (estxfer-xfer)/(1024*avgrate);
+		int hour = sec / 3600;
+		int min = (sec % 3600) / 60;
+		sec = sec % 60;
+		updgraph(0);
+		fprintf(file, "             %s %3i%%  ETA: %2i:%02i:%02i \n",
+			graph, (int)(100*xfer/estxfer), hour, min, sec);
+	} else
+		fprintf(file, "\n");
+
 }
 
 void printstatus(FILE* const file1, FILE* const file2,
@@ -183,17 +401,21 @@ void printstatus(FILE* const file1, FILE* const file2,
 {
 	float t1, t2; 
 	clock_t cl;
+	static int einvalwarn = 0;
 
 	if (file1 == stderr || file1 == stdout) 
-		fprintf(file1, "%s%s%s", up, up, up);
+		fprintf(file1, "%s%s%s%s", up, up, up, up);
 	if (file2 == stderr || file2 == stdout) 
-		fprintf(file2, "%s%s%s", up, up, up);
+		fprintf(file2, "%s%s%s%s", up, up, up, up);
 
 	if (sync) {
 		int err = fsync(odes);
-		if (err)
-			fplog(stderr, "dd_rescue: (warning): %s (%.1fk): %s!    \n", 
+		if (err && (errno != EINVAL || !einvalwarn) &&!o_chr) {
+			fplog(stderr, "dd_rescue: (warning): sync %s (%.1fk): %s!  \n",
 			      oname, (float)ipos/1024, strerror(errno));
+			++einvalwarn;
+		}
+		errno = 0;
 	}
 
 	gettimeofday(&currenttime, NULL);
@@ -205,20 +427,20 @@ void printstatus(FILE* const file1, FILE* const file2,
 		doprint(file1, bs, cl, t1, t2, sync);
 	if (file2)
 		doprint(file2, bs, cl, t1, t2, sync);
-	if (sync) {
+	if (1 || sync) {
 		memcpy(&lasttime, &currenttime, sizeof(lasttime));
 		lxfer = xfer;
 	}
 }
 
-void savebb( int block )
+static void savebb(int block)
 {
-	fplog(stderr, "Bad block: %d\n", block);
-    
-	if( bbname == NULL )
+	FILE *bbfile;
+	fplog(stderr, "Bad block reading %s: %d\n", iname, block);
+	if (bbname == NULL)
 		return;
-	FILE *bbfile = fopen(bbname,"a");
-	fprintf(bbfile,"%d\n",block);
+	bbfile = fopen(bbname, "a");
+	fprintf(bbfile, "%d\n", block);
 	fclose(bbfile);
 }
 
@@ -226,35 +448,59 @@ void printreport()
 {
 	/* report */
 	FILE *report = (!quiet || nrerr)? stderr: 0;
-	fplog(report, "Summary for %s -> %s:\n", iname, oname);
-	if (report)
-		fprintf(stderr, "%s%s%s", down, down, down);
-	printstatus(report, logfd, 0, 1);
+	if (report) {
+		fplog(report, "dd_rescue: (info): Summary for %s -> %s:\n", iname, oname);
+		fprintf(stderr, "%s%s%s%s", down, down, down, down);
+		printstatus(report, logfd, 0, 1);
+	}
+}
+
+static int mayexpandfile()
+{	
+	struct stat st;
+	off_t maxopos = opos;
+	if (init_opos > opos)
+		maxopos = init_opos;
+	stat(oname, &st);
+	if (!S_ISREG(st.st_mode))
+		return 0;
+	if (st.st_size < maxopos)
+		return truncate(oname, maxopos);
+	else 
+		return 0;		
 }
 
 int cleanup()
 {
 	int rc, errs = 0;
 	if (odes != -1) {
-		/* Make sure, the output file is expanded to the last (first) position */
+		/* Make sure, the output file is expanded to the last (first) position
+	 	 * FIXME: 0 byte writes do NOT expand file -- mayexpandfile() will
+		 * take care of that. */
 		pwrite(odes, buf, 0, opos);
 		rc = fsync(odes);
-		if (rc) {
-			fplog(stderr, "dd_rescue: (warning): %s (%.1fk): %s!\n", 
+		if (rc && !o_chr) {
+			fplog(stderr, "dd_rescue: (warning): fsync %s (%.1fk): %s!\n", 
 			      oname, (float)opos/1024, strerror(errno));
 			++errs;
+			errno = 0;
 		}
 		rc = close(odes); 
 		if (rc) {
-			fplog(stderr, "dd_rescue: (warning): %s (%.1fk): %s!\n", 
+			fplog(stderr, "dd_rescue: (warning): close %s (%.1fk): %s!\n", 
 			      oname, (float)opos/1024, strerror(errno));
 			++errs;
 		}
+		if (sparse)
+			rc = mayexpandfile();
+			if (rc)
+				fplog(stderr, "dd_rescue: (warning): seek %s (%1.fk): %s!\n",
+				      oname, (float)opos/1024, strerror(errno));
 	}
 	if (ides != -1) {
 		rc = close(ides);
 		if (rc) {
-			fplog(stderr, "dd_rescue: (warning): %s (%.1fk): %s!\n", 
+			fplog(stderr, "dd_rescue: (warning): close %s (%.1fk): %s!\n", 
 			      iname, (float)ipos/1024, strerror(errno));
 			++errs;
 		}
@@ -267,13 +513,13 @@ int cleanup()
 }
 
 /* is the block zero ? */
-int blockiszero(const char* blk, const int ln)
+static int blockiszero(const unsigned char* blk, const int ln)
 {
 	unsigned long* ptr = (unsigned long*)blk;
 	while ((ptr-(unsigned long*)blk) < ln/sizeof(unsigned long))
 		if (*(ptr++)) 
-			return 0;
-	return 1;
+			return (sizeof(unsigned long)*(ptr-(unsigned long*)blk));
+	return ln;
 }
 
 inline ssize_t mypread(int fd, void* bf, size_t sz, off_t off)
@@ -319,7 +565,7 @@ ssize_t writeblock(const int towrite)
 		  || (wr < towrite && err > 0 && errno == 0));
 	if (wr < towrite && err != 0) {
 		/* Write error: handle ? .. */
-		fplog(stderr, "dd_rescue: (%s): %s (%.1fk): %s\n",
+		fplog(stderr, "dd_rescue: (%s): write %s (%.1fk): %s\n",
 		      (abwrerr? "fatal": "warning"),
 		      oname, (float)opos/1024, strerror(errno));
 		if (abwrerr) {
@@ -330,185 +576,295 @@ ssize_t writeblock(const int towrite)
 	return (/*err == -1? err:*/ wr);
 }
 
-/* can be invoked in two ways: bs==hardbs or bs==softbs */
-int copyfile(const off_t max, const int bs)
+int blockxfer(const off_t max, const int bs)
 {
+	int block = bs;
+	/* Don't xfer more bytes than our limit */
+	if (max && max-xfer < bs)
+		block = max-xfer;
+	if (reverse) {
+		/* Can't go beyond the beginning of the file */
+		if (block > ipos)
+			block = ipos;
+		if (block > opos)
+			block = opos;
+	}
+	return block;
+}
+
+void exitfatalerr()
+{
+	if (errno == ESPIPE || errno == EPERM || errno == ENXIO || errno == ENODEV) {
+		fplog(stderr, "dd_rescue: (warning): %s (%.1fk): %s!\n", 
+		      iname, (float)ipos/1024, strerror(errno));
+		fplog(stderr, "dd_rescue: Last error fatal! Exiting ...\n");
+		cleanup();
+		exit(20);
+	}
+}
+
+
+static void advancepos(const ssize_t rd, const ssize_t wr)
+{
+	sxfer += wr; xfer += rd;
+	if (reverse) { 
+		ipos -= rd; opos -= rd; 
+	} else { 
+		ipos += rd; opos += rd; 
+	}
+}
+
+int dowrite(const ssize_t rd)
+{
+	int errs = 0;
+	int fatal = 0;
+	/* errno == 0: We can write to disk */
+	ssize_t wr = 0;
+	if (!sparse || blockiszero(buf, rd) < rd)
+		errs += ((wr = writeblock(rd)) < rd ? 1: 0);
+	advancepos(rd, wr);
+	if (wr <= 0 && (errno == ENOSPC 
+		   || (errno == EFBIG && !reverse)))
+		++fatal;
+	if (rd != wr && !sparse) {
+		fplog(stderr, "dd_rescue: (warning): assumption rd(%i) == wr(%i) failed! \n", rd, wr);
+		fplog(stderr, "dd_rescue: (%s): write %s (%.1fk): %s!\n", 
+		      (fatal? "fatal": "warning"), oname, 
+		      (float)opos/1024, strerror(errno));
+		fprintf(stderr, "%s%s%s", down, down, down);
+		errno = 0;
+	}
+	return fatal? -errs: errs;
+}
+
+static int partialwrite(const ssize_t rd)
+{
+	/* But first: write available data and advance (optimization) */
+	if (rd > 0 && !reverse) 
+		return dowrite(rd);
+	return 0;	
+}
+
+int copyfile_hardbs(const off_t max)
+{
+	ssize_t toread;
 	int errs = 0; errno = 0;
 #if 0	
-	fprintf(stderr, "%s%s%s copyfile (ipos=%.1fk, xfer=%.1fk, max=%.1fk, bs=%i)                         ##\n%s%s%s",
-		up, up, up,
-		(float)ipos/1024, (float)xfer/1024, (float)max/1024, bs,
-		down, down, down);
+	fprintf(stderr, "%s%s%s%s copyfile (ipos=%.1fk, xfer=%.1fk, max=%.1fk, bs=%i)                         ##\n%s%s%s%s",
+		up, up, up, up,
+		(float)ipos/1024, (float)xfer/1024, (float)max/1024, hardbs,
+		down, down, down, down);
 #endif
-	/* expand file to the right length */
-	if (!o_chr) 
-		pwrite(odes, buf, 0, opos);
-	while ( (!max || (max-xfer > 0))
-		&& ((!reverse) || (ipos > 0 && opos > 0)) ) {
-		int err;
-		ssize_t rd = 0;
-		ssize_t toread = ((max && max-xfer < bs)? (max-xfer): bs);
-		if (reverse) {
-			if (toread > ipos) 
-				toread = ipos;
-			if (toread > opos)
-				toread = opos;
-		}
-
-		if (nosparse && bs == hardbs)
-			memset(buf, 0, bs);
-		rd = readblock(toread);
+	while ((toread = blockxfer(max, hardbs)) > 0) { 
+		ssize_t rd = readblock(toread);
 
 		/* EOF */
 		if (rd == 0 && !errno) {
-			if (!errs)
-				fplog(stderr, "dd_rescue: (info): %s (%.1fk): EOF\n", 
+			if (!quiet)
+				fplog(stderr, "dd_rescue: (info): read %s (%.1fk): EOF\n", 
 				      iname, (float)ipos/1024);
 			return errs;
 		}
 		/* READ ERROR */
 		if (rd < toread/* && errno*/) {
-			/* Read error occurred: Print warning */
-			printstatus(stderr, logfd, bs, 1); 
-			errs++;
-			/* Some errnos are fatal */
-			if (errno == ESPIPE || errno == EPERM || errno == ENXIO || errno == ENODEV) {
-				fplog(stderr, "dd_rescue: (warning): %s (%.1fk): %s!\n", 
-				      iname, (float)ipos/1024, strerror(errno));
-				fplog(stderr, "dd_rescue: Last error fatal! Exiting ...\n");
-				cleanup(); exit(20);
+			if (errno) {
+				++errs;
+				/* Read error occurred: Print warning */
+				printstatus(stderr, logfd, hardbs, 1);
 			}
+			/* Some errnos are fatal */
+			exitfatalerr();
 			/* Non fatal error */
-			if (bs == hardbs) {
-				/* Real error: Don't retry */
-				nrerr++; 
-				fplog(stderr, "dd_rescue: (warning): %s (%.1fk): %s!\n", 
-				      iname, (float)ipos/1024, strerror(errno));
-				/* exit if too many errs */
-				if (maxerr && nrerr >= maxerr) {
-					fplog(stderr, "dd_rescue: (fatal): maxerr reached!\n");
-					printreport();
-					cleanup(); exit(32);
-				}
-				fprintf(stderr, "%s%s%s", down, down, down);
-				/* advance */
-				errno = 0;
-				if (nosparse) {
-					ssize_t wr = 0;
-					errs += ((wr = writeblock(rd)) < rd ? 1: 0);
-					if (wr < 0 && (errno == ENOSPC 
-						   || (errno == EFBIG && !reverse))) 
-						return errs;
-					if (rd != wr) {
-						fplog(stderr, "dd_rescue: (warning): assumption rd(%i) == wr(%i) failed! \n", rd, wr);	
-						/*
-						fplog(stderr, "dd_rescue: (warning): %s (%.1fk): %s!\n", 
-						      oname, (float)opos/1024, strerror(errno));
-						fprintf(stderr, "%s%s%s", down, down, down);
-					 	*/
-					}
-				}
-				savebb(ipos/bs);
-				fxfer += toread; xfer += toread;
-				if (reverse) { 
-					ipos -= toread; opos -= toread; 
-				} else { 
-					ipos += toread; opos += toread; 
-				}
-			} else {
-				off_t new_max = xfer + toread;
-				off_t old_xfer;
-				/* Error with large blocks: Try small ones ... */
-				if (verbose) 
-					fprintf(stderr, "dd_rescue: (info): problems at ipos %.1fk: %s \n                 fall back to smaller blocksize \n%s%s%s",
-					        (float)ipos/1024, strerror(errno), down, down, down);
-				/* But first: write available data and advance (optimization) */
-				if (rd > 0) {
-					ssize_t wr = 0; errno = 0;
-					if (!sparse || !blockiszero(buf, bs))
-						errs += ((wr = writeblock(rd)) < rd ? 1: 0);
-					if (!reverse) {
-						ipos += rd; opos += rd; 
-						sxfer += wr; xfer += rd;
-					}
-					/* 
-					else {
-						new_max -= rd;
-					}
-					 */
-					if (wr < 0 && (errno == ENOSPC 
-						   || (errno == EFBIG && !reverse))) 
-						return errs;
-					if (rd != wr && !sparse) {
-						fplog(stderr, "dd_rescue: (warning): assumption rd(%i) == wr(%i) failed! \n", rd, wr);
-						/*
-						fplog(stderr, "dd_rescue: (warning): %s (%.1fk): %s!\n", 
-						      oname, (float)opos/1024, strerror(errno));
-						fprintf(stderr, "%s%s%s", down, down, down);
-					 	 */
-					}
-				} /* rd > 0 */
-				old_xfer = xfer;
-				errs += (err = copyfile(new_max, hardbs));
-				/* EOF */
-				if (!err && old_xfer == xfer) 
-					return errs;
-				/*
-				if (reverse && rd) {
-					ipos -= rd; opos -= rd;
-					xfer += rd; sxfer += wr;
-				}
-				*/	
-				/* Stay with small blocks, until we could read two whole 
-				   large ones without errors */
-				new_max = xfer;
-				while (err && (!max || (max-xfer > 0)) && ((!reverse) || (ipos > 0 && opos > 0))) {
-					new_max += 2*softbs; old_xfer = xfer;
-					if (max && new_max > max) 
-						new_max = max;
-					errs += (err = copyfile(new_max, hardbs));
-				}
-				errno = 0;
-				/* EOF ? */      
-				if (!err && xfer == old_xfer) 
-					return errs;
-				if (verbose) 
-					fprintf(stderr, "dd_rescue: (info): ipos %.1fk promote to large bs again! \n%s%s%s",
-						(float)ipos/1024, down, down, down);
-			} /* bs == hardbs */
-		} else {
-	      		/* errno == 0: We can write to disk */
-      			if (rd > 0) {
+			/* Real error on small blocks: Don't retry */
+			nrerr++; 
+			fplog(stderr, "dd_rescue: (warning): read %s (%.1fk): %s!\n", 
+			      iname, (float)ipos/1024, strerror(errno));
+			/* exit if too many errs */
+			if (maxerr && nrerr >= maxerr) {
+				fplog(stderr, "dd_rescue: (fatal): maxerr reached!\n");
+				printreport();
+				cleanup(); exit(32);
+			}
+			fprintf(stderr, "%s%s%s%s", down, down, down, down);
+			
+			errno = 0;
+			if (nosparse || 
+			    (rd > 0 && (!sparse || blockiszero(buf, rd) < rd))) {
 				ssize_t wr = 0;
-				if (!sparse || !blockiszero(buf, bs))
-					errs += ((wr = writeblock(rd)) < rd ? 1: 0);
-				sxfer += wr; xfer += rd;
-				if (reverse) { 
-					ipos -= rd; opos -= rd; 
-				} else { 
-					ipos += rd; opos += rd; 
-				}
-				if (wr < 0 && (errno == ENOSPC 
+				memset(buf+rd, 0, toread-rd);
+				errs += ((wr = writeblock(toread)) < toread ? 1: 0);
+				if (wr <= 0 && (errno == ENOSPC 
 					   || (errno == EFBIG && !reverse))) 
 					return errs;
-				if (rd != wr && !sparse) {
-					fplog(stderr, "dd_rescue: (warning): assumption rd(%i) == wr(%i) failed! \n", rd, wr);
+				if (toread != wr) {
+					fplog(stderr, "dd_rescue: (warning): assumption toread(%i) == wr(%i) failed! \n", toread, wr);	
+					/*
 					fplog(stderr, "dd_rescue: (warning): %s (%.1fk): %s!\n", 
 					      oname, (float)opos/1024, strerror(errno));
-					fprintf(stderr, "%s%s", down, down);
-					errno = 0;
+					fprintf(stderr, "%s%s%s%s", down, down, down, down);
+				 	*/
 				}
-			} /* rd > 0 */
-		} /* errno */
+			}
+			savebb(ipos/hardbs);
+			updgraph(1);
+			fxfer += toread; xfer += toread;
+			if (reverse) { 
+				ipos -= toread; opos -= toread; 
+			} else { 
+				ipos += toread; opos += toread; 
+			}
+		} else {
+	      		int err = dowrite(rd);
+			if (err < 0)
+				return -err;
+			else
+				errs += err;
+		}
 
 		if (syncfreq && !(xfer % (syncfreq*softbs)))
-			printstatus((quiet? 0: stderr), 0, bs, 1);
+			printstatus((quiet? 0: stderr), 0, hardbs, 1);
 		else if (!quiet && !(xfer % (16*softbs)))
-			printstatus(stderr, 0, bs, 0);
+			printstatus(stderr, 0, hardbs, 0);
 	} /* remain */
 	return errs;
 }
 
+int copyfile_softbs(const off_t max)
+{
+	ssize_t toread;
+	int errs = 0; errno = 0;
+#if 0	
+	fprintf(stderr, "%s%s%s%s copyfile (ipos=%.1fk, xfer=%.1fk, max=%.1fk, bs=%i)                         ##\n%s%s%s%s",
+		up, up, up, up,
+		(float)ipos/1024, (float)xfer/1024, (float)max/1024, softbs,
+		down, down, down, down);
+#endif
+	/* expand file to AT LEAST the right length 
+	 * FIXME: 0 byte writes do NOT expand file */
+	if (!o_chr)
+		pwrite(odes, buf, 0, opos);
+	while ((toread = blockxfer(max, softbs)) > 0) { 
+		int err;
+		ssize_t rd = readblock(toread);
+
+		/* EOF */
+		if (rd == 0 && !errno) {
+			if (!quiet)
+				fplog(stderr, "dd_rescue: (info): read %s (%.1fk): EOF\n", 
+				      iname, (float)ipos/1024);
+			return errs;
+		}
+		/* READ ERROR or short read */
+		if (rd < toread/* && errno*/) {
+			int ret;
+			off_t new_max, old_xfer;
+			if (errno) {
+				++errs;
+				/* Read error occurred: Print warning */
+				printstatus(stderr, logfd, softbs, 1);
+			}
+			/* Some errnos are fatal */
+			exitfatalerr();
+			/* Non fatal error */
+			new_max = xfer + toread;
+			/* Error with large blocks: Try small ones ... */
+			if (verbose) {
+				fprintf(stderr, "dd_rescue: (info): problems at ipos %.1fk: %s \n                 fall back to smaller blocksize \n%s%s%s%s",
+				        (float)ipos/1024, strerror(errno), down, down, down, down);
+				printstatus(stderr, logfd, hardbs, 1);
+			}
+			/* But first: write available data and advance (optimization) */
+			if ((ret = partialwrite(rd)) < 0)
+				return ret;
+			else
+				errs += ret;
+			old_xfer = xfer;
+			errs += (err = copyfile_hardbs(new_max));
+			/* EOF */
+			if (!err && old_xfer == xfer)
+				return errs;
+			/*
+			if (reverse && rd) {
+				ipos -= rd; opos -= rd;
+				xfer += rd; sxfer += wr;
+			}
+			*/	
+			/* Stay with small blocks, until we could read two whole 
+			   large ones without errors */
+			new_max = xfer;
+			while (err && (!max || (max-xfer > 0)) && ((!reverse) || (ipos > 0 && opos > 0))) {
+				new_max += 2*softbs; old_xfer = xfer;
+				if (max && new_max > max) 
+					new_max = max;
+				errs += (err = copyfile_hardbs(new_max));
+			}
+			errno = 0;
+			/* EOF ? */      
+			if (!err && xfer == old_xfer)
+				return errs;
+			if (verbose) 
+				fprintf(stderr, "dd_rescue: (info): ipos %.1fk promote to large bs again! \n%s%s%s%s",
+					(float)ipos/1024, down, down, down, down);
+		} else {
+	      		err = dowrite(rd);
+			if (err < 0)
+				return -err;
+			else
+				errs += err;
+		} /* errno */
+
+		if (syncfreq && !(xfer % (syncfreq*softbs)))
+			printstatus((quiet? 0: stderr), 0, softbs, 1);
+		else if (!quiet && !(xfer % (16*softbs)))
+			printstatus(stderr, 0, softbs, 0);
+	} /* remain */
+	return errs;
+}
+
+#ifdef HAVE_SPLICE
+int copyfile_splice(const off_t max)
+{
+	ssize_t toread;
+	int fd_pipe[2];
+	if (pipe(fd_pipe) < 0)
+		return copyfile_softbs(max);
+	while ((toread	= blockxfer(max, softbs)) > 0) {
+		ssize_t rd = splice(ides, &ipos, fd_pipe[1], NULL, toread,
+					SPLICE_F_MOVE | SPLICE_F_MORE);
+		if (rd < 0) {
+			close(fd_pipe[0]); close(fd_pipe[1]);
+			if (!quiet)
+				fplog(stderr, "dd_rescue: (info): %s (%.1fk): fall back to userspace copy\n%s%s%s%s", 
+				      iname, (float)ipos/1024, down, down, down, down);
+			return copyfile_softbs(max);
+		}
+		if (rd == 0) {
+			if (!quiet)
+				fplog(stderr, "dd_rescue: (info): read %s (%.1fk): EOF (splice)\n", 
+				      iname, (float)ipos/1024);
+			close(fd_pipe[0]); close(fd_pipe[1]);
+			return 0;
+		}
+		while (rd) {
+			ssize_t wr = splice(fd_pipe[0], NULL, odes, &opos, rd,
+					SPLICE_F_MOVE | SPLICE_F_MORE);
+			if (wr < 0) {
+				close(fd_pipe[0]); close(fd_pipe[1]);
+				exit(23);
+			}
+			rd -= wr; xfer += wr; sxfer += wr;
+		}
+		advancepos(0, 0);
+		if (syncfreq && !(xfer % (syncfreq*softbs)))
+			printstatus((quiet? 0: stderr), 0, softbs, 1);
+		else if (!quiet && !(xfer % (16*softbs)))
+			printstatus(stderr, 0, softbs, 0);
+	}
+	close(fd_pipe[0]); close(fd_pipe[1]);
+	return 0;
+}
+#endif
+
 int copyperm(int ifd, int ofd)
 {
 	int err; 
@@ -540,7 +896,7 @@ int copytimes(const char* inm, const char* onm)
 	return err;
 }
 
-off_t readint(const char* const ptr)
+static off_t readint(const char* const ptr)
 {
 	char *es; double res;
 
@@ -562,6 +918,20 @@ void printversion()
 {
 	fprintf(stderr, "\ndd_rescue Version %s, garloff@suse.de, GNU GPL\n", VERSION);
 	fprintf(stderr, " (%s)\n", ID);
+	fprintf(stderr, " (compiled %s %s by %s)\n", __DATE__, __TIME__, __COMPILER__);
+	fprintf(stderr, " (features: ");
+#ifdef O_DIRECT
+	fprintf(stderr, "O_DIRECT ");
+#endif
+#ifdef HAVE_LIBFALLOCATE
+	fprintf(stderr, "libfallocate ");
+#elif defined(HAVE_FALLOCATE)
+	fprintf(stderr, "fallocate ");
+#endif
+#ifdef HAVE_SPLICE
+	fprintf(stderr, "splice ");
+#endif
+	fprintf(stderr, ")\n");
 }
 
 void printhelp()
@@ -582,6 +952,12 @@ void printhelp()
 	fprintf(stderr, "         -t         truncate output file (def=no),\n");
 #ifdef O_DIRECT
 	fprintf(stderr, "         -d/D       use O_DIRECT for input/output (def=no),\n");
+#endif
+#ifdef HAVE_SPLICE
+	fprintf(stderr, "         -k         use efficient in-kernel zerocopy splice\n");
+#endif       	
+#ifdef HAVE_FALLOCATE
+	fprintf(stderr, "         -P         use fallocate to preallocate target space\n");
 #endif
 	fprintf(stderr, "         -w         abort on Write errors (def=no),\n");
 	fprintf(stderr, "         -a         spArse file writing (def=no),\n");
@@ -633,7 +1009,9 @@ int main(int argc, char* argv[])
 {
 	int c;
 	off_t syncsz = -1;
+#ifdef O_DIRECT
 	void **mp = (void **) &buf;
+#endif
 
   	/* defaults */
 	softbs = SOFTBLOCKSIZE; hardbs = HARDBLOCKSIZE;
@@ -641,13 +1019,14 @@ int main(int argc, char* argv[])
 	reverse = 0; dotrunc = 0; abwrerr = 0; sparse = 0; nosparse = 0;
 	verbose = 0; quiet = 0; interact = 0; force = 0; pres = 0;
 	lname = 0; iname = 0; oname = 0; o_dir_in = 0; o_dir_out = 0;
+	dosplice = 0; falloc = 0;
 
 	/* Initialization */
 	sxfer = 0; fxfer = 0; lxfer = 0; xfer = 0;
 	ides = -1; odes = -1; logfd = 0; nrerr = 0; buf = 0;
 	i_chr = 0; o_chr = 0;
 
-	while ((c = getopt(argc, argv, ":rtfihqvVwaAdDpb:B:m:e:s:S:l:o:y:")) != -1) {
+	while ((c = getopt(argc, argv, ":rtfihqvVwaAdDkpPb:B:m:e:s:S:l:o:y:")) != -1) {
 		switch (c) {
 			case 'r': reverse = 1; break;
 			case 't': dotrunc = O_TRUNC; break;
@@ -657,7 +1036,11 @@ int main(int argc, char* argv[])
 			case 'd': o_dir_in  = O_DIRECT; break;
 			case 'D': o_dir_out = O_DIRECT; break;
 #endif
+#ifdef HAVE_SPLICE
+			case 'k': dosplice = 1; break;
+#endif				  
 			case 'p': pres = 1; break;
+			case 'P': falloc = 1; break;
 			case 'a': sparse = 1; nosparse = 0; break;
 			case 'A': nosparse = 1; sparse = 0; break;
 			case 'w': abwrerr = 1; break;
@@ -685,6 +1068,7 @@ int main(int argc, char* argv[])
 		}
 	}
   
+	init_opos = opos;
 	if (optind < argc) 
 		iname = argv[optind++];
 	if (optind < argc) 
@@ -730,9 +1114,10 @@ int main(int argc, char* argv[])
 	}
 
 	/* Set sync frequency */
+	/*
 	if (syncsz == -1)
 		syncfreq = 512;
-	else if (syncsz == 0)
+	else */ if (syncsz <= 0)
 		syncfreq = 0;
 	else
 		syncfreq = (syncsz + softbs - 1) / softbs;
@@ -766,7 +1151,7 @@ int main(int argc, char* argv[])
 	if (ides < 0) {
 		fplog(stderr, "dd_rescue: (fatal): %s: %s\n", iname, strerror(errno));
 		cleanup(); exit(22);
-	};
+	}
 
 	/* Overwrite? */
 	/* Special case '-': stdout */
@@ -783,7 +1168,7 @@ int main(int argc, char* argv[])
 		do {
 			fprintf(stderr, "dd_rescue: (question): %s existing %s [y/n] ?", 
 				(dotrunc? "Overwrite": "Write into"), oname);
-			a = toupper(fgetc (stdin)); //fprintf(stderr, "\n");
+			a = toupper(fgetc(stdin)); //fprintf(stderr, "\n");
 		} while (a != 'Y' && a != 'N');
 		if (a == 'N') {
 			fplog(stderr, "dd_rescue: (fatal): exit on user request!\n");
@@ -801,10 +1186,12 @@ int main(int argc, char* argv[])
 		copyperm(ides, odes);
 			
 	check_seekable(ides, odes);
-	if (0 && i_chr && o_chr) {
-		fprintf(stderr, "dd_rescue: (fatal): Sorry, there is no support yet for non-seekable\n");
-		fprintf(stderr, "                    input and output. This will hopefully change soon ... \n");
-		exit(19);
+
+	sparse_output_warn();
+	if (o_chr) {
+		if (!nosparse)
+			fprintf(stderr, "dd_rescue: (warning): Don't use sparse writes for non-seekable output\n");
+		nosparse = 1; sparse = 0; dosplice = 0;
 	}
 
 	/* special case: reverse with ipos == 0 means ipos = end_of_file */
@@ -861,6 +1248,21 @@ int main(int argc, char* argv[])
 		cleanup(); exit(19);
 	}
 		
+	if (dosplice) {
+		if (!quiet)
+			fplog(stderr, "dd_rescue: (info): splice copy, ignoring -a, -r, -y\n");
+		reverse = 0;
+	}
+
+	input_length();
+#if 0
+	fplog(stderr, "dd_rescue: (info): copy %Li bytes from file %s (%Li) to %s\n",
+		estxfer, iname, ilen, oname);
+#endif
+#ifdef HAVE_FALLOCATE
+	if (falloc)
+		do_fallocate();
+#endif
 
 	if (verbose) {
 		printinfo(stderr);
@@ -880,11 +1282,21 @@ int main(int argc, char* argv[])
 	memcpy(&lasttime, &starttime, sizeof(lasttime));
 
 	if (!quiet) {
-		fprintf(stderr, "%s%s%s", down, down, down);
+		fprintf(stderr, "%s%s%s%s", down, down, down, down);
 		printstatus(stderr, 0, softbs, 0);
 	}
 
-	c = copyfile(maxxfer, softbs);
+#ifdef HAVE_SPLICE
+	if (dosplice)
+		c = copyfile_splice(maxxfer);
+	else 
+#endif
+	{
+		if (softbs > hardbs)
+			c = copyfile_softbs(maxxfer);
+		else
+			c = copyfile_hardbs(maxxfer);
+	}
 
 	gettimeofday(&currenttime, NULL);
 	printreport();
